/*
   Copyright (c) 2009-2011, Jack Poulson
   All rights reserved.

   This file is part of Elemental-FLAME.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

    - Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    - Neither the name of the owner nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/
#include "elemental.hpp"
using namespace std;
using namespace elemental;
using namespace elemental::imports;

// Typedef our real and complex types to 'R' and 'C' for convenience
typedef double R;
typedef complex<R> C;

int
main( int argc, char* argv[] )
{
    // Initialize Elemental. The full routine is elemental::Init.
    // It detects whether or not you have already initialized MPI.
    Init( argc, argv );

    // Extract our MPI rank
    mpi::Comm comm = mpi::COMM_WORLD;
    const int rank = mpi::CommRank( comm );

    // Surround the Elemental calls with try/catch statements in order to 
    // safely handle any exceptions that were thrown during execution.
    try 
    {
        // Create a 2d process grid from a communicator. In our case, it is
        // MPI_COMM_WORLD. There is another constructor that allows you to 
        // specify the grid dimensions, Grid g( comm, r, c ), which creates an 
        // r x c grid.
        Grid g( comm );
    
        // Create an n x n complex distributed matrix, 
        // [MC,MR] is standard 2d matrix distribution. 
        // We distribute the matrix using grid 'g'.
        //
        // There are quite a few available constructors, including ones that 
        // allow you to pass in your own local buffer and to specify the 
        // distribution alignments (i.e., which process row and column owns the
        // top-left element)
        const int n = 6; // choose a small problem size since we will print
        DistMatrix<C,MC,MR> H( n, n, g );

        // Fill the matrix since we did not pass in a buffer. 
        //
        // We will fill entry (i,j) with the complex value (i+j,i-j) so that 
        // the global matrix is Hermitian. However, only one triangle of the 
        // matrix actually needs to be filled, the symmetry can be implicit.
        //
        const int r = g.Height(); // height of 2d process grid
        const int c = g.Width(); // width of 2d process grid
        const int colShift = H.ColShift(); // first row we own
        const int rowShift = H.RowShift(); // first col we own
        const int localHeight = H.LocalHeight();
        const int localWidth = H.LocalWidth();
        for( int jLocal=0; jLocal<localWidth; ++jLocal )
        {
            for( int iLocal=0; iLocal<localHeight; ++iLocal )
            {
                // Our process owns the rows { n \in N_0 : n mod r = colShift }
                //              and the cols { n \in N_0 : n mod c = rowShift }
                const int i = colShift + iLocal*r;
                const int j = rowShift + jLocal*c;
                H.SetLocalEntry( iLocal, jLocal, C(i+j,i-j) );
            }
        }
        // Alternatively, we could have sequentially filled the matrix with 
        // for( int j=0; j<A.Width(); ++j )
        //   for( int i=0; i<A.Height(); ++i )
        //     A.Set( i, j, C(i+j,i-j) );
        //
        // More convenient interfaces are being investigated.
        //

        // Print our matrix.
        H.Print("H:");

        // Call the eigensolver. We first create an empty complex eigenvector 
        // matrix, X[MC,MR], and an eigenvalue row vector, w[* ,VR]
        DistMatrix<R,Star,VR> w( g );
        DistMatrix<C,MC,MR> X( g );
        // Optional: set blocksizes and algorithmic choices here. See the 
        //           'Tuning' section of the README for details.
        advanced::HermitianEig( Lower, H, w, X ); // only use lower half of H

        // Print the eigensolution
        w.Print("Eigenvalues of H:");
        X.Print("Eigenvectors of H:");
    }
    catch( exception& e )
    {
        cerr << "Process " << rank << " caught exception with message: "
             << e.what() << endl;
    }

    Finalize();
    return 0;
}

