/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2011, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "blis.h"

void bli_strmm( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, float* alpha, float* a, int a_rs, int a_cs, float* b, int b_rs, int b_cs )
{
	int       m_save    = m;
	int       n_save    = n;
	float*    a_save    = a;
	float*    b_save    = b;
	int       a_rs_save = a_rs;
	int       a_cs_save = a_cs;
	int       b_rs_save = b_rs;
	int       b_cs_save = b_cs;
	int       dim_a;
	int       lda, inca;
	int       ldb, incb;

	// Return early if possible.
	if ( bli_zero_dim2( m, n ) ) return;

	// If necessary, allocate, initialize, and use a temporary contiguous
	// copy of each matrix rather than the original matrices.
	bli_set_dim_with_side( side, m, n, &dim_a );
	bli_screate_contigmr( uplo,
	                      dim_a,
	                      dim_a,
	                      a_save, a_rs_save, a_cs_save,
	                      &a,     &a_rs,     &a_cs );

	bli_screate_contigm( m,
	                     n,
	                     b_save, b_rs_save, b_cs_save,
	                     &b,     &b_rs,     &b_cs );

	// Initialize with values assuming column-major storage.
	lda  = a_cs;
	inca = a_rs;
	ldb  = b_cs;
	incb = b_rs;

	// Adjust the parameters based on the storage of each matrix.
	if ( bli_is_col_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr( uplo( A_c ) ) * B_c
			// effective operation: B_c := tr( uplo( A_c ) ) * B_c
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr(  uplo( A_r ) )   * B_c
			// effective operation: B_c := tr( ~uplo( A_c ) )^T * B_c
			bli_swap_ints( lda, inca );

			bli_toggle_uplo( uplo );
			bli_toggle_trans( trans );
		}
	}
	else // if ( bli_is_row_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_c ) ) * B_r
			// effective operation: B_c := B_c * tr( uplo( A_c ) )^T
			bli_swap_ints( ldb, incb );

			bli_swap_ints( m, n );

			bli_toggle_side( side );
			bli_toggle_trans( trans );
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_r ) ) * B_r
			// effective operation: B_c := B_c * tr( ~uplo( A_c ) )
			bli_swap_ints( ldb, incb );
			bli_swap_ints( lda, inca );

			bli_swap_ints( m, n );

			bli_toggle_uplo( uplo );
			bli_toggle_side( side );
		}
	}

	bli_strmm_blas( side,
	                uplo,
	                trans,
	                diag,
	                m,
	                n,
	                alpha,
	                a, lda,
	                b, ldb );

	// Free any temporary contiguous matrices, copying the result back to
	// the original matrix.
	bli_sfree_contigm( a_save, a_rs_save, a_cs_save,
	                   &a,     &a_rs,     &a_cs );

	bli_sfree_saved_contigm( m_save,
	                         n_save,
	                         b_save, b_rs_save, b_cs_save,
	                         &b,     &b_rs,     &b_cs );
}

void bli_dtrmm( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, double* alpha, double* a, int a_rs, int a_cs, double* b, int b_rs, int b_cs )
{
	int       m_save    = m;
	int       n_save    = n;
	double*   a_save    = a;
	double*   b_save    = b;
	int       a_rs_save = a_rs;
	int       a_cs_save = a_cs;
	int       b_rs_save = b_rs;
	int       b_cs_save = b_cs;
	int       dim_a;
	int       lda, inca;
	int       ldb, incb;

	// Return early if possible.
	if ( bli_zero_dim2( m, n ) ) return;

	// If necessary, allocate, initialize, and use a temporary contiguous
	// copy of each matrix rather than the original matrices.
	bli_set_dim_with_side( side, m, n, &dim_a );
	bli_dcreate_contigmr( uplo,
	                      dim_a,
	                      dim_a,
	                      a_save, a_rs_save, a_cs_save,
	                      &a,     &a_rs,     &a_cs );

	bli_dcreate_contigm( m,
	                     n,
	                     b_save, b_rs_save, b_cs_save,
	                     &b,     &b_rs,     &b_cs );

	// Initialize with values assuming column-major storage.
	lda  = a_cs;
	inca = a_rs;
	ldb  = b_cs;
	incb = b_rs;

	// Adjust the parameters based on the storage of each matrix.
	if ( bli_is_col_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr( uplo( A_c ) ) * B_c
			// effective operation: B_c := tr( uplo( A_c ) ) * B_c
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr(  uplo( A_r ) )   * B_c
			// effective operation: B_c := tr( ~uplo( A_c ) )^T * B_c
			bli_swap_ints( lda, inca );

			bli_toggle_uplo( uplo );
			bli_toggle_trans( trans );
		}
	}
	else // if ( bli_is_row_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_c ) ) * B_r
			// effective operation: B_c := B_c * tr( uplo( A_c ) )^T
			bli_swap_ints( ldb, incb );

			bli_swap_ints( m, n );

			bli_toggle_side( side );
			bli_toggle_trans( trans );
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_r ) ) * B_r
			// effective operation: B_c := B_c * tr( ~uplo( A_c ) )
			bli_swap_ints( ldb, incb );
			bli_swap_ints( lda, inca );

			bli_swap_ints( m, n );

			bli_toggle_uplo( uplo );
			bli_toggle_side( side );
		}
	}

	bli_dtrmm_blas( side,
	                uplo,
	                trans,
	                diag,
	                m,
	                n,
	                alpha,
	                a, lda,
	                b, ldb );

	// Free any temporary contiguous matrices, copying the result back to
	// the original matrix.
	bli_dfree_contigm( a_save, a_rs_save, a_cs_save,
	                   &a,     &a_rs,     &a_cs );

	bli_dfree_saved_contigm( m_save,
	                         n_save,
	                         b_save, b_rs_save, b_cs_save,
	                         &b,     &b_rs,     &b_cs );
}

void bli_ctrmm( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, scomplex* alpha, scomplex* a, int a_rs, int a_cs, scomplex* b, int b_rs, int b_cs )
{
	int       m_save    = m;
	int       n_save    = n;
	scomplex* a_save    = a;
	scomplex* b_save    = b;
	int       a_rs_save = a_rs;
	int       a_cs_save = a_cs;
	int       b_rs_save = b_rs;
	int       b_cs_save = b_cs;
	scomplex* a_conj;
	int       dim_a;
	int       lda, inca;
	int       ldb, incb;
	int       lda_conj, inca_conj;
	int       a_was_copied;

	// Return early if possible.
	if ( bli_zero_dim2( m, n ) ) return;

	// If necessary, allocate, initialize, and use a temporary contiguous
	// copy of each matrix rather than the original matrices.
	bli_set_dim_with_side( side, m, n, &dim_a );
	bli_ccreate_contigmr( uplo,
	                      dim_a,
	                      dim_a,
	                      a_save, a_rs_save, a_cs_save,
	                      &a,     &a_rs,     &a_cs );

	bli_ccreate_contigm( m,
	                     n,
	                     b_save, b_rs_save, b_cs_save,
	                     &b,     &b_rs,     &b_cs );

	// Figure out whether A was copied to contiguous memory. This is used to
	// prevent redundant copying.
	a_was_copied = ( a != a_save );

	// Initialize with values assuming column-major storage.
	lda  = a_cs;
	inca = a_rs;
	ldb  = b_cs;
	incb = b_rs;

	// Adjust the parameters based on the storage of each matrix.
	if ( bli_is_col_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr( uplo( A_c ) ) * B_c
			// effective operation: B_c := tr( uplo( A_c ) ) * B_c
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr(  uplo( A_r ) )   * B_c
			// effective operation: B_c := tr( ~uplo( A_c ) )^T * B_c
			bli_swap_ints( lda, inca );

			bli_toggle_uplo( uplo );
			bli_toggle_trans( trans );
		}
	}
	else // if ( bli_is_row_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_c ) ) * B_r
			// effective operation: B_c := B_c * tr( uplo( A_c ) )^T
			bli_swap_ints( ldb, incb );

			bli_swap_ints( m, n );

			bli_toggle_side( side );
			bli_toggle_trans( trans );
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_r ) ) * B_r
			// effective operation: B_c := B_c * tr( ~uplo( A_c ) )
			bli_swap_ints( ldb, incb );
			bli_swap_ints( lda, inca );

			bli_swap_ints( m, n );

			bli_toggle_uplo( uplo );
			bli_toggle_side( side );
		}
	}

	// Initialize with values assuming that trans is not conjnotrans.
	a_conj    = a;
	lda_conj  = lda;
	inca_conj = inca;

	// We want to handle the conjnotrans case. The easiest way to do so is
	// by making a conjugated copy of A.
	if ( bli_is_conjnotrans( trans ) && !a_was_copied )
	{
		int dim_a;

		bli_set_dim_with_side( side, m, n, &dim_a );
		
		a_conj    = bli_callocm( dim_a, dim_a );
		lda_conj  = dim_a;
		inca_conj = 1;

		bli_ccopymrt( uplo,
		              BLIS_CONJ_NO_TRANSPOSE,
		              dim_a,
		              dim_a,
		              a,      inca,      lda,
		              a_conj, inca_conj, lda_conj );
	}
	else if ( bli_is_conjnotrans( trans ) && a_was_copied )
	{
		int dim_a;

		bli_set_dim_with_side( side, m, n, &dim_a );
		
		bli_cconjmr( uplo,
		             dim_a,
		             dim_a,
		             a_conj, inca_conj, lda_conj );
	}


	bli_ctrmm_blas( side,
	                uplo,
	                trans,
	                diag,
	                m,
	                n,
	                alpha,
	                a_conj, lda_conj,
	                b,      ldb );

	if ( bli_is_conjnotrans( trans ) && !a_was_copied )
		bli_cfree( a_conj );

	// Free any temporary contiguous matrices, copying the result back to
	// the original matrix.
	bli_cfree_contigm( a_save, a_rs_save, a_cs_save,
	                   &a,     &a_rs,     &a_cs );

	bli_cfree_saved_contigm( m_save,
	                         n_save,
	                         b_save, b_rs_save, b_cs_save,
	                         &b,     &b_rs,     &b_cs );
}

void bli_ztrmm( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, dcomplex* alpha, dcomplex* a, int a_rs, int a_cs, dcomplex* b, int b_rs, int b_cs )
{
	int       m_save    = m;
	int       n_save    = n;
	dcomplex* a_save    = a;
	dcomplex* b_save    = b;
	int       a_rs_save = a_rs;
	int       a_cs_save = a_cs;
	int       b_rs_save = b_rs;
	int       b_cs_save = b_cs;
	dcomplex* a_conj;
	int       dim_a;
	int       lda, inca;
	int       ldb, incb;
	int       lda_conj, inca_conj;
	int       a_was_copied;

	// Return early if possible.
	if ( bli_zero_dim2( m, n ) ) return;

	// If necessary, allocate, initialize, and use a temporary contiguous
	// copy of each matrix rather than the original matrices.
	bli_set_dim_with_side( side, m, n, &dim_a );
	bli_zcreate_contigmr( uplo,
	                      dim_a,
	                      dim_a,
	                      a_save, a_rs_save, a_cs_save,
	                      &a,     &a_rs,     &a_cs );

	bli_zcreate_contigm( m,
	                     n,
	                     b_save, b_rs_save, b_cs_save,
	                     &b,     &b_rs,     &b_cs );

	// Figure out whether A was copied to contiguous memory. This is used to
	// prevent redundant copying.
	a_was_copied = ( a != a_save );

	// Initialize with values assuming column-major storage.
	lda  = a_cs;
	inca = a_rs;
	ldb  = b_cs;
	incb = b_rs;

	// Adjust the parameters based on the storage of each matrix.
	if ( bli_is_col_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr( uplo( A_c ) ) * B_c
			// effective operation: B_c := tr( uplo( A_c ) ) * B_c
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_c := tr(  uplo( A_r ) )   * B_c
			// effective operation: B_c := tr( ~uplo( A_c ) )^T * B_c
			bli_swap_ints( lda, inca );

			bli_toggle_uplo( uplo );
			bli_toggle_trans( trans );
		}
	}
	else // if ( bli_is_row_storage( b_rs, b_cs ) )
	{
		if ( bli_is_col_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_c ) ) * B_r
			// effective operation: B_c := B_c * tr( uplo( A_c ) )^T
			bli_swap_ints( ldb, incb );

			bli_swap_ints( m, n );

			bli_toggle_side( side );
			bli_toggle_trans( trans );
		}
		else // if ( bli_is_row_storage( a_rs, a_cs ) )
		{
			// requested operation: B_r := tr( uplo( A_r ) ) * B_r
			// effective operation: B_c := B_c * tr( ~uplo( A_c ) )
			bli_swap_ints( ldb, incb );
			bli_swap_ints( lda, inca );

			bli_swap_ints( m, n );

			bli_toggle_side( side );
			bli_toggle_uplo( uplo );
		}
	}

	// Initialize with values assuming that trans is not conjnotrans.
	a_conj    = a;
	lda_conj  = lda;
	inca_conj = inca;

	// We want to handle the conjnotrans case. The easiest way to do so is
	// by making a conjugated copy of A.
	if ( bli_is_conjnotrans( trans ) && !a_was_copied )
	{
		int dim_a;

		bli_set_dim_with_side( side, m, n, &dim_a );
		
		a_conj    = bli_zallocm( dim_a, dim_a );
		lda_conj  = dim_a;
		inca_conj = 1;

		bli_zcopymrt( uplo,
		              BLIS_CONJ_NO_TRANSPOSE,
		              dim_a,
		              dim_a,
		              a,      inca,      lda,
		              a_conj, inca_conj, lda_conj );
	}
	else if ( bli_is_conjnotrans( trans ) && a_was_copied )
	{
		int dim_a;

		bli_set_dim_with_side( side, m, n, &dim_a );
		
		bli_zconjmr( uplo,
		             dim_a,
		             dim_a,
		             a_conj, inca_conj, lda_conj );
	}

	bli_ztrmm_blas( side,
	                uplo,
	                trans,
	                diag,
	                m,
	                n,
	                alpha,
	                a_conj, lda_conj,
	                b,      ldb );

	if ( bli_is_conjnotrans( trans ) && !a_was_copied )
		bli_zfree( a_conj );

	// Free any temporary contiguous matrices, copying the result back to
	// the original matrix.
	bli_zfree_contigm( a_save, a_rs_save, a_cs_save,
	                   &a,     &a_rs,     &a_cs );

	bli_zfree_saved_contigm( m_save,
	                         n_save,
	                         b_save, b_rs_save, b_cs_save,
	                         &b,     &b_rs,     &b_cs );
}

// --- Classic routine wrappers ---

void bli_strmm_blas( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, float* alpha, float* a, int lda, float* b, int ldb )
{
#ifdef BLIS_ENABLE_CBLAS_INTERFACES
	enum CBLAS_ORDER cblas_order = CblasColMajor;
	enum CBLAS_SIDE  cblas_side;
	enum CBLAS_UPLO  cblas_uplo;
	enum CBLAS_TRANSPOSE cblas_trans;
	enum CBLAS_DIAG  cblas_diag;

	bli_param_map_to_netlib_side( side, &cblas_side );
	bli_param_map_to_netlib_uplo( uplo, &cblas_uplo );
	bli_param_map_to_netlib_trans( trans, &cblas_trans );
	bli_param_map_to_netlib_diag( diag, &cblas_diag );

	cblas_strmm( cblas_order,
	             cblas_side,
	             cblas_uplo,
	             cblas_trans,
	             cblas_diag,
	             m,
	             n,
	             *alpha,
	             a, lda,
	             b, ldb );
#else
	char blas_side;
	char blas_uplo;
	char blas_trans;
	char blas_diag;

	bli_param_map_to_netlib_side( side, &blas_side );
	bli_param_map_to_netlib_uplo( uplo, &blas_uplo );
	bli_param_map_to_netlib_trans( trans, &blas_trans );
	bli_param_map_to_netlib_diag( diag, &blas_diag );

	F77_strmm( &blas_side,
	           &blas_uplo,
	           &blas_trans,
	           &blas_diag,
	           &m,
	           &n,
	           alpha,
	           a, &lda,
	           b, &ldb );
#endif
}

void bli_dtrmm_blas( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, double* alpha, double* a, int lda, double* b, int ldb )
{
#ifdef BLIS_ENABLE_CBLAS_INTERFACES
	enum CBLAS_ORDER cblas_order = CblasColMajor;
	enum CBLAS_SIDE  cblas_side;
	enum CBLAS_UPLO  cblas_uplo;
	enum CBLAS_TRANSPOSE cblas_trans;
	enum CBLAS_DIAG  cblas_diag;

	bli_param_map_to_netlib_side( side, &cblas_side );
	bli_param_map_to_netlib_uplo( uplo, &cblas_uplo );
	bli_param_map_to_netlib_trans( trans, &cblas_trans );
	bli_param_map_to_netlib_diag( diag, &cblas_diag );

	cblas_dtrmm( cblas_order,
	             cblas_side,
	             cblas_uplo,
	             cblas_trans,
	             cblas_diag,
	             m,
	             n,
	             *alpha,
	             a, lda,
	             b, ldb );
#else
	char blas_side;
	char blas_uplo;
	char blas_trans;
	char blas_diag;

	bli_param_map_to_netlib_side( side, &blas_side );
	bli_param_map_to_netlib_uplo( uplo, &blas_uplo );
	bli_param_map_to_netlib_trans( trans, &blas_trans );
	bli_param_map_to_netlib_diag( diag, &blas_diag );

	F77_dtrmm( &blas_side,
	           &blas_uplo,
	           &blas_trans,
	           &blas_diag,
	           &m,
	           &n,
	           alpha,
	           a, &lda,
	           b, &ldb );
#endif
}

void bli_ctrmm_blas( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, scomplex* alpha, scomplex* a, int lda, scomplex* b, int ldb )
{
#ifdef BLIS_ENABLE_CBLAS_INTERFACES
	enum CBLAS_ORDER cblas_order = CblasColMajor;
	enum CBLAS_SIDE  cblas_side;
	enum CBLAS_UPLO  cblas_uplo;
	enum CBLAS_TRANSPOSE cblas_trans;
	enum CBLAS_DIAG  cblas_diag;

	bli_param_map_to_netlib_side( side, &cblas_side );
	bli_param_map_to_netlib_uplo( uplo, &cblas_uplo );
	bli_param_map_to_netlib_trans( trans, &cblas_trans );
	bli_param_map_to_netlib_diag( diag, &cblas_diag );

	cblas_ctrmm( cblas_order,
	             cblas_side,
	             cblas_uplo,
	             cblas_trans,
	             cblas_diag,
	             m,
	             n,
	             alpha,
	             a, lda,
	             b, ldb );
#else
	char blas_side;
	char blas_uplo;
	char blas_trans;
	char blas_diag;

	bli_param_map_to_netlib_side( side, &blas_side );
	bli_param_map_to_netlib_uplo( uplo, &blas_uplo );
	bli_param_map_to_netlib_trans( trans, &blas_trans );
	bli_param_map_to_netlib_diag( diag, &blas_diag );

	F77_ctrmm( &blas_side,
	           &blas_uplo,
	           &blas_trans,
	           &blas_diag,
	           &m,
	           &n,
	           alpha,
	           a, &lda,
	           b, &ldb );
#endif
}

void bli_ztrmm_blas( side_t side, uplo_t uplo, trans_t trans, diag_t diag, int m, int n, dcomplex* alpha, dcomplex* a, int lda, dcomplex* b, int ldb )
{
#ifdef BLIS_ENABLE_CBLAS_INTERFACES
	enum CBLAS_ORDER cblas_order = CblasColMajor;
	enum CBLAS_SIDE  cblas_side;
	enum CBLAS_UPLO  cblas_uplo;
	enum CBLAS_TRANSPOSE cblas_trans;
	enum CBLAS_DIAG  cblas_diag;

	bli_param_map_to_netlib_side( side, &cblas_side );
	bli_param_map_to_netlib_uplo( uplo, &cblas_uplo );
	bli_param_map_to_netlib_trans( trans, &cblas_trans );
	bli_param_map_to_netlib_diag( diag, &cblas_diag );

	cblas_ztrmm( cblas_order,
	             cblas_side,
	             cblas_uplo,
	             cblas_trans,
	             cblas_diag,
	             m,
	             n,
	             alpha,
	             a, lda,
	             b, ldb );
#else
	char blas_side;
	char blas_uplo;
	char blas_trans;
	char blas_diag;

	bli_param_map_to_netlib_side( side, &blas_side );
	bli_param_map_to_netlib_uplo( uplo, &blas_uplo );
	bli_param_map_to_netlib_trans( trans, &blas_trans );
	bli_param_map_to_netlib_diag( diag, &blas_diag );

	F77_ztrmm( &blas_side,
	           &blas_uplo,
	           &blas_trans,
	           &blas_diag,
	           &m,
	           &n,
	           alpha,
	           a, &lda,
	           b, &ldb );
#endif
}

