/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2011, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Apply_GTG( FLA_Obj gamma, FLA_Obj sigma,
                         FLA_Obj delta1, FLA_Obj epsilon, FLA_Obj delta2 )
/*
  Apply a Givens rotation

    G  =  / gamma  -sigma \
          \ sigma   gamma /

  to the left and the right of a 2x2 real symmetric matrix:

    / delta1   epsilon \  :=  G' / delta1   epsilon \ G
    \ epsilon  delta2  /         \ epsilon  delta2  /

  -FGVZ
*/
{
	FLA_Datatype datatype;

	datatype = FLA_Obj_datatype( delta1 );

	if ( FLA_Check_error_level() >= FLA_MIN_ERROR_CHECKING )
		FLA_Apply_GTG_check( gamma, sigma, delta1, epsilon, delta2 );

	switch ( datatype )
	{
		case FLA_FLOAT:
		{
			float* gamma_p   = ( float* ) FLA_FLOAT_PTR( gamma );
			float* sigma_p   = ( float* ) FLA_FLOAT_PTR( sigma );
			float* delta1_p  = ( float* ) FLA_FLOAT_PTR( delta1 );
			float* epsilon_p = ( float* ) FLA_FLOAT_PTR( epsilon );
			float* delta2_p  = ( float* ) FLA_FLOAT_PTR( delta2 );

			FLA_Apply_GTG_ops( gamma_p,
			                   sigma_p,
			                   delta1_p,
			                   epsilon_p,
			                   delta2_p );

			break;
		}

		case FLA_DOUBLE:
		{
			double* gamma_p   = ( double* ) FLA_FLOAT_PTR( gamma );
			double* sigma_p   = ( double* ) FLA_FLOAT_PTR( sigma );
			double* delta1_p  = ( double* ) FLA_FLOAT_PTR( delta1 );
			double* epsilon_p = ( double* ) FLA_FLOAT_PTR( epsilon );
			double* delta2_p  = ( double* ) FLA_FLOAT_PTR( delta2 );

			FLA_Apply_GTG_opd( gamma_p,
			                   sigma_p,
			                   delta1_p,
			                   epsilon_p,
			                   delta2_p );

			break;
		}
	}

	return FLA_SUCCESS;
}



FLA_Error FLA_Apply_GTG_ops( float*  gamma,
                             float*  sigma,
                             float*  delta1,
                             float*  epsilon,
                             float*  delta2 )
{
	return FLA_SUCCESS;
}

FLA_Error FLA_Apply_GTG_opd( double* gamma,
                             double* sigma,
                             double* delta1,
                             double* epsilon,
                             double* delta2 )
{
	double gamma_delta1;
	double gamma_epsilon;
	double gamma_delta2;
	double sigma_delta1;
	double sigma_epsilon;
	double sigma_delta2;
	double delta1_new;
	double epsilon_new;
	double delta2_new;
	double temp1, temp2;

	// Pre-compute some terms.

	gamma_delta1  = (*gamma) * (*delta1);
	gamma_epsilon = (*gamma) * (*epsilon);
	gamma_delta2  = (*gamma) * (*delta2);

	sigma_delta1  = (*sigma) * (*delta1);
	sigma_epsilon = (*sigma) * (*epsilon);
	sigma_delta2  = (*sigma) * (*delta2);

	// Compute new value for delta1.

	delta1_new = ( gamma_delta1  + sigma_epsilon ) * (*gamma) +
	             ( gamma_epsilon + sigma_delta2  ) * (*sigma);

	// Compute new values for epsilon and delta2.

	temp1 = -sigma_delta1  + gamma_epsilon;
	temp2 = -sigma_epsilon + gamma_delta2;
	
	epsilon_new = temp1 *  (*gamma) + temp2 * (*sigma);
	delta2_new  = temp1 * -(*sigma) + temp2 * (*gamma);

	// Overwrite the original values with the results.

	*delta1  = delta1_new;
	*epsilon = epsilon_new;
	*delta2  = delta2_new;

/*
	double d2_minus_d1; 
	double two_sigma_eps; 
	double d1, e1, d2; 
	double temp, temp2; 

	d2_minus_d1   = *(delta2) - *(delta1); 
	two_sigma_eps = 2.0 * *(sigma) * *(epsilon); 

	temp   = *(gamma) * d2_minus_d1 - two_sigma_eps; 
	e1     = *(sigma) * temp + *(epsilon); 
	temp2  = *(gamma) * temp; 
	d1     = -temp2 + *(delta2); 
	d2     =  temp2 + *(delta1);

	*(delta1)  = d1; 
	*(epsilon) = e1; 
	*(delta2)  = d2; 
*/

	return FLA_SUCCESS;
}

